"""Module containing the class implementation of a rule.

Class: Rule
"""

from __future__ import annotations

import re
from typing import Any, Callable

from arta.condition import BaseCondition, StandardCondition
from arta.exceptions import ConditionExecutionError, RuleExecutionError
from arta.utils import (
    ParsingErrorStrategy,
    parse_dynamic_parameter,
)


class Rule:
    """A rule is the combination of some conditions and one action.

    Attributes:
        set_id: An id of a rule set.
        group_id: An id of a rule group.
        rule_id: The id of the rule.
        condition_exprs: A dictionary of condition expressions (k: condition conf. key, v: condition expression).
        condition_factory_mapping: A dictionary mapping between condition conf. keys and condition class objects
                                    (k: condition conf. key, v: condition class object).
        action: Function to perform when the conditions are valid (action function).
        action_parameters: Parameters of the action function.
    """

    def __init__(
        self,
        set_id: str,
        group_id: str,
        rule_id: str,
        condition_exprs: dict[str, str | None],
        condition_factory_mapping: dict[str, type[BaseCondition]],
        action: Callable,
        std_condition_instances: dict[str, StandardCondition],
        action_parameters: dict[str, Any] | None = None,
    ) -> None:
        """Initialize attributes.

        Args:
            std_condition_instances: Dictionary containing the BaseCondition instances required by the Rule
                (k: condition_id, v: StandardCondition instance).
        """
        # IDs
        self._set_id = set_id
        self._group_id = group_id
        self._rule_id = rule_id

        # Action
        self._action = action
        self._action_parameters = action_parameters or {}

        # Condition expressions
        self._condition_exprs = condition_exprs

        # Factory mapping
        self._condition_factory_mapping = condition_factory_mapping

        # Condition instances (k: condition id (not conf key), v: instances)
        self._condition_instances: dict[str, BaseCondition] = self._instantiate_conditions(std_condition_instances)

    def apply(
        self,
        input_data: dict[str, Any],
        *,
        parsing_error_strategy: ParsingErrorStrategy,
        **kwargs: Any,
    ) -> tuple[Any | None, dict[str, Any]]:
        """Apply the rule on the input data, return action output (optional).

        Args:
            input_data: Request or input data to apply rules on.
            parsing_error_strategy: Parsing error strategy.
            **kwargs: For user extra arguments.

        Returns:
            A tuple as: (action result, rule result details).

        Raises:
            RuleExecutionError: Error during the rule execution.
        """
        # If rule conditions are verified, the action is executed w/ the parameters' value
        is_conditions_ok: bool
        rule_results: dict[str, Any]

        is_conditions_ok, rule_results = self._check_conditions(
            input_data, parsing_error_strategy=parsing_error_strategy, **kwargs
        )

        if is_conditions_ok:
            try:
                # Parse dynamic parameters
                parameters: dict[str, Any] = {}
                for key, value in self._action_parameters.items():
                    parameters[key] = parse_dynamic_parameter(
                        parameter=value, input_data=input_data, parsing_error_strategy=parsing_error_strategy
                    )

                # Track the rule id
                rule_results["activated_rule"] = self._rule_id

                # Run action
                rule_results["action_result"] = self._action(**parameters, input_data=input_data)

                return rule_results["action_result"], rule_results
            except Exception as error:
                raise RuleExecutionError(f"Error while executing rule '{self._rule_id}': {str(error)}") from error

        else:
            return None, {}

    def _check_conditions(
        self, input_data: dict[str, Any], parsing_error_strategy: ParsingErrorStrategy, **kwargs: Any
    ) -> tuple[bool, dict[str, Any]]:
        """(Protected)
        Return True if all conditions are verified.

        Args:
            input_data: Request or input data to apply rules on.
            parsing_error_strategy: Error handling strategy for parameter's parsing.
            **kwargs: For user extra arguments.

        Returns:
            A tuple as: (True if all conditions are verified, otherwise False, condition results dictionary).
        """
        # Var init.
        all_conditions_res: bool = True
        condition_results: dict[str, Any] = {"rule_group": self._group_id, "verified_conditions": {}}

        # Loop among condition expressions
        for cond_conf_key, expr in self._condition_exprs.items():
            condition_class: type[BaseCondition] = self._condition_factory_mapping[cond_conf_key]

            # Evaluate the condition expression
            try:
                condition_res, unitary_res = self._evaluate_condition_expr(
                    input_data=input_data,
                    condition_class=condition_class,
                    condition_expr=expr,
                    parsing_error_strategy=parsing_error_strategy,
                    **kwargs,
                )
            except NameError as e:
                raise RuleExecutionError(f"Error during evaluation of '{cond_conf_key}: {expr}': {str(e)}") from e

            # Combine conditions (AND)
            all_conditions_res = all_conditions_res and condition_res

            # Store condition results
            condition_results["verified_conditions"].update(
                {cond_conf_key: {"expression": expr, "values": unitary_res}}
            )

            if not all_conditions_res:
                # If False, no need to go further
                break

        return all_conditions_res, condition_results

    def _evaluate_condition_expr(
        self,
        input_data: dict[str, Any],
        condition_class: type[BaseCondition],
        parsing_error_strategy: ParsingErrorStrategy,
        condition_expr: str | None = None,
        **kwargs: Any,
    ) -> tuple[bool, dict[str, bool]]:
        """(Protected)
        Evaluate the condition expr (a boolean expression) and
        return the result (a boolean).

        Args:
            input_data: Request or input data.
            condition_class: Class object of the analyzed condition (given by its conf. key).
            parsing_error_strategy: Error handling strategy for parameter's parsing.
            condition_expr: A boolean expression (string).
            **kwargs: For user extra arguments.

        Returns:
            A tuple as: (final result, unitary results (dictionary)).

        Raises:
            ConditionExecutionError: Error during condition execution.
        """
        # Var init.
        unitary_results: dict[str, bool] = {}

        # Case of null condition expressions => Always True
        if condition_expr is None:
            return True, unitary_results

        # The final boolean expression is formed by replacing condition IDs by their evaluated boolean values
        bool_expr: str = condition_expr

        # Extract condition ids from the expression
        condition_ids: set[str] = condition_class.extract_condition_ids_from_expression(condition_expr)

        # Loop among the conditions of the expression
        # Verify the unitary condition
        for cond_id in condition_ids:
            # Retrieve condition instance
            condition: BaseCondition = self._condition_instances[cond_id]

            # Check unitary condition
            try:
                bool_var: bool = condition.verify(input_data, parsing_error_strategy=parsing_error_strategy, **kwargs)

                # Store unitary result
                unitary_results[cond_id] = bool_var
            except Exception as error:
                raise ConditionExecutionError(f"Error while executing condition '{cond_id}': {str(error)}") from error

            # Replace the result in the boolean expression
            sanit_cond_id: str = condition.get_sanitized_id()
            bool_expr = re.sub(rf"{sanit_cond_id}", str(bool_var), bool_expr)

        # Evaluate the final boolean expressions = final result
        return eval(bool_expr), unitary_results  # noqa

    def _instantiate_conditions(
        self,
        std_conditions: dict[str, StandardCondition],
    ) -> dict[str, BaseCondition]:
        """Parse condition expressions and build corresponding instances.

        E.g., for one condition:
        - Input : "not(CONDITION_A) and CONDITION_B"
        - Output : {"CONDITION_A": condition_A_instance, "CONDITION_B": condition_B_instance}

        Args:
            std_conditions: A dictionary containing the StandardCondition instances
                (k: cond. id, v: StandardCondition instance)

        Returns:
            Condition instances which are in the condition expressions (k: condition id, v: BaseCondition instance).

        Raises:
            KeyError: Condition id is unknown.
        """
        # Var init.
        condition_ids: set[str] = set()
        cond_instances: dict[str, BaseCondition] = {}

        # Get all condition ids from the expressions (1 or many)
        for conf_key, expr in self._condition_exprs.items():
            # Expression parsing is condition class dependent
            condition_ids = self._condition_factory_mapping[conf_key].extract_condition_ids_from_expression(expr)

            # Is a custom condition or a simple condition?
            if self._condition_factory_mapping is not None and conf_key != "condition":
                # Yes
                for cond_id in condition_ids:
                    # Instanciate the custom (unknown) condition object
                    cond_instances[cond_id] = self._condition_factory_mapping[conf_key](
                        condition_id=cond_id,
                        description=self._condition_factory_mapping[conf_key].CONDITION_DATA_LABEL,
                    )
            else:
                # Should be a standard condition
                for cond_id in condition_ids:
                    # Retrieve the standard condition instance (already instantiated)
                    try:
                        cond_instances[cond_id] = std_conditions[cond_id]
                    except KeyError as error:
                        raise KeyError(
                            f"Following condition id is unknown '{cond_id}' in {conf_key}: {expr}"
                        ) from error

        return cond_instances



"""Configuration handling module."""

from __future__ import annotations

from pathlib import Path
from typing import Any, cast

from omegaconf import DictConfig, ListConfig, OmegaConf


def load_config(config_dir_path: str) -> dict[str, Any]:
    """Load a configuration dictionary from all the yaml files in a given directory (and its subdirectories).

    Args:
        config_dir_path: Path to a directory containing YML files.
        prefix: Prefix for the rglob pattern.
        exclude_pattern: Regex pattern to exclude files.

    Returns:
        config: Loaded config dictionary.
    """
    conf_files: list[Path] = [f for patt in ["*.yml", "*.yaml"] for f in Path(config_dir_path).rglob(patt)]

    # Alphabetical sorting of the file names, it enables splitting the rule sets over different files
    conf_files.sort()

    omega_config: DictConfig | ListConfig = OmegaConf.unsafe_merge(*[OmegaConf.load(file) for file in conf_files])
    config: dict[str, Any] = cast(dict[str, Any], OmegaConf.to_object(omega_config))

    return config



"""Condition implementation.

Classes: BaseCondition, StandardCondition, SimpleCondition
"""

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from typing import Any, Callable

from arta.exceptions import ConditionExecutionError
from arta.utils import ParsingErrorStrategy, parse_dynamic_parameter


class BaseCondition(ABC):
    """Base class of a Condition object (Strategy Pattern).

    Is an abstract class and can't be instantiated.

    Attributes:
        condition_id: Id of a condition.
        description: Description of a condition.
        validation_function: Validation function of a condition.
        validation_function_parameters: Arguments of the validation function.
    """

    # Class constants
    CONDITION_DATA_LABEL: str = "Undefined condition data (not needed)"
    CONDITION_ID_PATTERN: str = r"\b[A-Z_0-9]+\b"

    def __init__(
        self,
        condition_id: str,
        description: str,
        validation_function: Callable | None = None,
        validation_function_parameters: dict[str, Any] | None = None,
    ) -> None:
        """
        Initialize attributes.

        Args:
            condition_id: Id of a condition.
            description: Description of a condition.
            validation_function: Validation function of a condition.
            validation_function_parameters: Arguments of the validation function.
        """
        self._condition_id = condition_id
        self._description = description
        self._validation_function = validation_function
        self._validation_function_parameters = validation_function_parameters

    @abstractmethod
    def verify(self, input_data: dict[str, Any], parsing_error_strategy: ParsingErrorStrategy, **kwargs: Any) -> bool:
        """(Abstract)
        Return True if the condition is verified.

        Args:
            input_data: Input data to apply rules on.
            parsing_error_strategy: Error handling strategy for parameter parsing.
            **kwargs: For user extra arguments.

        Returns:
            True if the condition is verified, otherwise False.
        """
        raise NotImplementedError

    def get_sanitized_id(self) -> str:
        """Return the sanitized (regex) condition id.

        E.g., 'CONDITION_2'       --> '\\bCONDITION_2\\b'

        Returns:
            A sanitized regex pattern string.
        """
        return rf"\b{self._condition_id}\b"

    @classmethod
    def extract_condition_ids_from_expression(cls, condition_expr: str | None = None) -> set[str]:
        """Get the condition ids from a string (e.g., UPPERCASE words).

        E.g., CONDITION_1 and not CONDITION_2

        Warning: implementation is based on the current class constant CONDITION_SPLIT_PATTERN.

        Args:
            condition_expr: A boolean expression (string).

        Returns:
            A set of extracted condition ids.
        """
        cond_ids: set[str] = set()

        if condition_expr is not None:
            cond_ids = set(re.findall(cls.CONDITION_ID_PATTERN, condition_expr))

        return cond_ids


class StandardCondition(BaseCondition):
    """Class implementing a built-in condition, named standard condition.

    Attributes:
        condition_id: Id of a condition.
        description: Description of a condition.
        validation_function: Validation function of a condition.
        validation_function_parameters: Arguments of the validation function.
    """

    # Class constants
    CONDITION_DATA_LABEL: str = "Standard condition (will be overwritten)"

    def verify(self, input_data: dict[str, Any], parsing_error_strategy: ParsingErrorStrategy, **kwargs: Any) -> bool:
        """Return True if the condition is verified.

        Example of a unitary standard condition: CONDITION_1

        Args:
            input_data: Request or input data to apply rules on.
            parsing_error_strategy: Error handling strategy for parameter parsing.
            **kwargs: For user extra arguments.

        Returns:
            True if the condition is verified, otherwise False.

        Raises:
            AttributeError: Check the validation function or its parameters.
        """
        if self._validation_function is None:
            raise AttributeError("Validation function should not be None")

        if self._validation_function_parameters is None:
            raise AttributeError("Validation function parameters should not be None")

        # Parse dynamic parameters
        parameters: dict[str, Any] = {}

        for key, value in self._validation_function_parameters.items():
            parameters[key] = parse_dynamic_parameter(
                parameter=value, input_data=input_data, parsing_error_strategy=parsing_error_strategy
            )

        # Run validation_function
        return self._validation_function(**parameters)


class SimpleCondition(BaseCondition):
    """Class implementing a built-in simple condition.

    Attributes:
        condition_id: Id of a condition.
        description: Description of a condition.
        validation_function: Validation function of a condition.
        validation_function_parameters: Arguments of the validation function.
    """

    # Class constants
    CONDITION_DATA_LABEL: str = "Simple condition data (not needed)"
    CONDITION_ID_PATTERN: str = r"(?:input\.|output\.)(?:[a-zA-Z0-9!=<>\"NTF\.\*\+\-_/]*)(?:[a-zA-Z\s\-_]*\"|)"

    def verify(self, input_data: dict[str, Any], parsing_error_strategy: ParsingErrorStrategy, **kwargs: Any) -> bool:
        """Return True if the condition is verified.

        Example of a unitary simple condition to be verified: 'input.age>=100'

        Args:
            input_data: Request or input data to apply rules on.
            parsing_error_strategy: Error handling strategy for parameter parsing.
            **kwargs: For user extra arguments.

        Returns:
            True if the condition is verified, otherwise False.

        Raises:
            AttributeError: Check the validation function or its parameters.
        """
        bool_var: bool = False
        unitary_expr: str = self._condition_id

        data_path_patt: str = r"(?:input\.|output\.)(?:[a-zA-Z_\.]*)"

        # Retrieve only the data path
        path_matches: list[str] = re.findall(data_path_patt, unitary_expr)

        if len(path_matches) > 0:
            locals_ns: dict[str, Any] = {}

            # Regular case: we have a data paths
            for idx, path in enumerate(path_matches):
                # Read data from the path
                locals_ns[f"data_{idx}"] = parse_dynamic_parameter(
                    parameter=path, input_data=input_data, parsing_error_strategy=parsing_error_strategy
                )

                # Replace with the variable name in the expression
                unitary_expr = unitary_expr.replace(path, f"data_{idx}")

            # Evaluate the expression
            try:
                bool_var = eval(unitary_expr, None, locals_ns)  # noqa
            except TypeError:
                # Ignore evaluation --> False
                pass

        elif parsing_error_strategy == ParsingErrorStrategy.RAISE:
            # Raise an error because of no match for a data path
            raise ConditionExecutionError(f"Error when verifying simple condition: '{unitary_expr}'")

        else:
            # Other case: ignore, default value => return False
            pass

        return bool_var

    def get_sanitized_id(self) -> str:
        """Return the sanitized (regex) condition id.

        E.g., 'input.power=="fly"' --> 'input\\.power==\\"fly\\"'

        Returns:
            A sanitized regex pattern string.
        """
        return re.escape(self._condition_id)



"""Utility module."""

from __future__ import annotations

import copy
import re
from enum import Enum
from typing import Any


class ParsingErrorStrategy(str, Enum):
    """Define authorized error handling strategies when a key is missing in the input data."""

    RAISE: str = "raise"
    IGNORE: str = "ignore"
    DEFAULT_VALUE: str = "default_value"


class RuleActivationMode(str, Enum):
    """Define how Arta is processing rules.

    ONE_BY_GROUP is the default mode.
    """

    ONE_BY_GROUP: str = "one_by_group"
    MANY_BY_GROUP: str = "many_by_group"


def get_value_in_nested_dict_from_path(path: str, nested_dict: dict[str, Any]) -> Any:
    """From a path, get a value in a nested dict.

    Ex:
        path : this.is.a.path
        result : nested_dict["this"]["is"]["a"]["path"]

    Args:
        path: A dictionary path.
        nested_dict: A nested dictionary.

    Returns:
        Found value.
    """
    keys: list[str] = path.split(".")

    # Initialize value with whole nested dict
    value: dict[str, Any] = nested_dict

    # Loop on path keys
    for key in keys:
        if value is None:
            raise KeyError(f"Key {value} of path {path} not found in input data.")
        value = value[key]

    return value


def parse_dynamic_parameter(
    parameter: Any,
    input_data: dict[str, Any],
    parsing_error_strategy: ParsingErrorStrategy,
) -> Any:
    """Parse the value of parameterized parameters.

    (e.g.1, input.age  -> 20, e.g.2, input.name.first -> "John")

    Args:
        parameter: The parameters configured in conditions.yaml.
        input_data: Request or input data to apply rules on.
        parsing_error_strategy: Strategy to adopt when confronted with a missing key.

    Returns:
        The list of the parameters' values.

    Raises:
        KeyError: Key not found.
    """
    # Copy parameters to not alterate original
    parameter = copy.deepcopy(parameter)

    if isinstance(parameter, list):
        parameter = [parse_dynamic_parameter(element, input_data, parsing_error_strategy) for element in parameter]

    elif isinstance(parameter, str):
        if not parameter.startswith(("input.", "output.")):
            # Keep parameter value unchanged
            return parameter

        # Remove the "input" prefix
        param_path: str = re.sub(r"^input\.", r"", parameter)

        # Check if a parsing error strategy flag is present
        default_value, param_path, parsing_error_strategy = check_parsing_error_strategy_override(
            param_path, parsing_error_strategy
        )

        # Get value from path
        try:
            parameter = get_value_in_nested_dict_from_path(path=param_path, nested_dict=input_data)
        except KeyError as error:
            if parsing_error_strategy is ParsingErrorStrategy.IGNORE:
                return None
            if parsing_error_strategy is ParsingErrorStrategy.DEFAULT_VALUE:
                return default_value
            else:
                raise KeyError(f"Could not find path '{param_path}' in the input data: {str(error)}") from error

    return parameter


def check_parsing_error_strategy_override(
    param_path: str, parsing_error_strategy: ParsingErrorStrategy
) -> tuple[Any, str, ParsingErrorStrategy]:
    """Check if the input parameter contains a flag to override the parsing error strategy.

    The following override syntaxes are accepted:
    - output.favorite_meal!             # raise an exception
    - output.favorite_meal?             # parameter = None
    - output.favorite_meal?default_str  # parameter = default_str (works on str only at first)

    Args:
        param_path: Path to a parameter in a nested dict.
        parsing_error_strategy: Strategy to adopt when a missing key occured.

    Returns:
        default_value: Default value to use if the DEFAULT_VALUE strategy is adopted.
        param_path: Clean path to the input parameter.
        parsing_error_strategy: Strategy adopted to handle parsing errors.
    """
    # Override default error handling strategy
    default_value: Any = None

    last_key: str = param_path.split(".")[-1]

    # Replace missing fields with None
    if last_key.endswith("?"):
        # Set strategy
        parsing_error_strategy = ParsingErrorStrategy.IGNORE
        # Cleanup path
        param_path = param_path.rstrip("?")

    # Raise error on missing fields
    elif last_key.endswith("!"):
        # Set strategy
        parsing_error_strategy = ParsingErrorStrategy.RAISE
        # Cleanup path
        param_path = param_path.rstrip("!")

    # Replace missing fields with default value
    elif "?" in last_key:
        default_value = last_key.split("?")[-1]
        # Set strategy
        parsing_error_strategy = ParsingErrorStrategy.DEFAULT_VALUE
        # Cleanup path
        param_path = "".join(param_path.split("?")[:-1])

    return default_value, param_path, parsing_error_strategy
